package at.tugraz.kmi.medokyservice.fca.db;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import at.tugraz.kmi.medokyservice.fca.db.conf.DBConfig;
import at.tugraz.kmi.medokyservice.fca.db.domainmodel.FCAAttribute;
import at.tugraz.kmi.medokyservice.fca.db.domainmodel.FCAObject;
import at.tugraz.kmi.medokyservice.fca.rest.FCAService;

public class MySQLConnector {

  private static final String   PREFIX_STUDENT                 = "stud";
  private static final String   PREFIX_ACTIVITY                = "act";
  private static final String   PREFIX_COMPETENCE              = "comp";

  // Defaults in case no conifg has been provided
  private static final String   DB_URL                         = "localhost:3306/mylea_beta";
  private static final String   DB_UNAME                       = "root";
  private static final String   DB_PWD                         = "root";

  // statements (See DB Schema)
  private static final String   STMT_STUDENTS                  = "SELECT people.id,people.firstname,people.surname FROM people INNER JOIN group_has_students ON people.id=group_has_students.studentid WHERE groupid=?";
  private static final String   STMT_GROUP_NAME                = "SELECT name FROM groups WHERE id=?";
  private static final String   STMT_SUBJECTS_BY_GROUP         = "SELECT subjectid FROM group_has_subjects WHERE groupid=?";
  private static final String   STMT_SUBJECT_NAMES             = "SELECT id,name FROM subjects";
  private static final String   STMT_COMPETENCIES_BY_GROUP     = "SELECT competency_data.compid, competencies.name FROM competency_data INNER JOIN competencies ON competency_data.compid=competencies.id WHERE groupid=?";
  private static final String   STMT_COMPETENCIES_BY_SUBJECT   = "SELECT subject_has_competencies.subjectid, subject_has_competencies.compid, competencies.name FROM subject_has_competencies INNER JOIN competencies ON subject_has_competencies.compid=competencies.id WHERE subject_has_competencies.subjectid IN(#)";
  private static final String   STMT_ACTIVITIES_BY_SUBJECT     = "SELECT activities_has_subjects.activityid, activities.name, activities_has_subjects.subjectid, activities.threshold FROM activities_has_subjects INNER JOIN activities ON activities.id=activities_has_subjects.activityid WHERE activities_has_subjects.subjectid IN (#)";
  private static final String   STMT_COMPLOG                   = "SELECT * FROM (SELECT studid,compid,value,timestamp FROM complog where studid IN ($) AND compid IN (#) ORDER BY timestamp DESC LIMIT 18446744073709551615) as t1 GROUP BY compid,studid";
  private static final String   STMT_ACTLOG                    = "SELECT * FROM (SELECT studentid,activityid,value,timestamp FROM activitieslog where studentid IN ($) AND activityid IN (#) ORDER BY timestamp DESC LIMIT 18446744073709551615) as t1 GROUP BY activityid,studentid";
  private static final String   STMT_ACTIVITIES_BY_COMPETENCES = "SELECT activities_has_competencies.activityid, activities.name, activities_has_competencies.competenceid FROM activities_has_competencies INNER JOIN activities ON activities.id=activities_has_competencies.activityid INNER JOIN activities_has_subjects ON activities_has_competencies.activityid=activities_has_subjects.activityid WHERE activities_has_subjects.subjectid IN (#)";

  private static MySQLConnector instance;

  private final Properties      dbProps;
  private Connection            conn;

  private MySQLConnector() throws ClassNotFoundException, SQLException, IOException {
    Class.forName("com.mysql.jdbc.Driver");
    // Create directories
    Database.getInstance();

    // read MySQL config
    dbProps = readConfig();

    // connect
    connect();

  }

  private void connect() throws SQLException {
    conn = DriverManager.getConnection("jdbc:mysql://" + dbProps.getProperty(DBConfig.KEY_MYSQL_URL, DB_URL),
        dbProps.getProperty(DBConfig.KEY_MYSQL_USER, DB_UNAME), dbProps.getProperty(DBConfig.KEY_MYSQL_PWD, DB_PWD));
  }

  /**
   * Read SQL URL+Credentials
   * 
   * @return
   * @throws IOException
   */
  private Properties readConfig() throws IOException {
    Properties props = new Properties();
    File propFile = new File(DBConfig.MYSQL_CONF_PATH);
    if (propFile.exists()) {
      FCAService.log("Reading properties from " + propFile.getCanonicalPath());
      props.load(new FileInputStream(propFile));
      FCAService.log("MySQL URL: " + props.getProperty(DBConfig.KEY_MYSQL_URL, DB_URL));
      FCAService.log("MySQL USR: " + props.getProperty(DBConfig.KEY_MYSQL_USER, DB_UNAME));
    } else
      FCAService.log("Configuration " + propFile.getCanonicalPath() + " does not exist. using defaults");
    return props;
  }

  public static MySQLConnector getInstance() throws ClassNotFoundException, SQLException, IOException {
    if (MySQLConnector.instance == null) {
      MySQLConnector.instance = new MySQLConnector();
    }
    synchronized (instance) {
      if (!instance.conn.isValid(2))
        instance.connect();
    }
    return MySQLConnector.instance;
  }

  /**
   * Get all students of a group
   * 
   * @param gid
   *          the group's ID
   * @return mapping of Student MySQL DB IDs to Students as FCA Attributes
   * @throws SQLException
   */
  public Map<Integer, FCAAttribute> getStudents(final int gid) throws SQLException {
    final PreparedStatement stmt = conn.prepareStatement(MySQLConnector.STMT_STUDENTS);
    stmt.setInt(1, gid);
    final ResultSet rsStudents = stmt.executeQuery();
    final Map<Integer, FCAAttribute> students = new HashMap<>();
    while (rsStudents.next()) {
      final String id = MySQLConnector.PREFIX_STUDENT + rsStudents.getInt(1);
      students.put(rsStudents.getInt(1),
          new FCAAttribute(rsStudents.getString(2) + " " + rsStudents.getString(3), "", id));
    }
    return students;
  }

  /**
   * Get a group's name
   * 
   * @param gid
   *          the group's ID
   * @return the group's name
   * @throws SQLException
   */
  public String getGroupName(final int gid) throws SQLException {
    final PreparedStatement stmt = conn.prepareStatement(MySQLConnector.STMT_GROUP_NAME);
    stmt.setInt(1, gid);
    final ResultSet rsGName = stmt.executeQuery();
    if (!rsGName.next()) {
      return null;
    }
    return rsGName.getString(1);
  }
  //
  // public Map<String, String> getSubjectNames() throws SQLException {
  // final Statement stmt = conn.createStatement();
  // final ResultSet rsSubjectNames = stmt.executeQuery(MySQLConnector.STMT_SUBJECT_NAMES);
  // final Map<String, String> subjectNames = new HashMap<>();
  // while (rsSubjectNames.next()) {
  // subjectNames.put(rsSubjectNames.getString(1), rsSubjectNames.getString(2));
  // }
  // return subjectNames;
  // }

  /**
   * Get all activities associated to a group
   * 
   * @param gid
   *          the group's ID
   * @return mapping of activities' MySQL DB IDs to activities as FCA Objects with their threshold values set into the
   *         description field
   * @throws SQLException
   */
  public Map<Integer, FCAObject> getActivites(int gid) throws SQLException {
    final Set<Integer> subjectIDsByGroup = getSubjectIDsByGroup(gid);
    final String qryActivities = MySQLConnector.buildStmtIN(MySQLConnector.STMT_ACTIVITIES_BY_SUBJECT, "#",
        subjectIDsByGroup.size());

    final PreparedStatement stmtActivites = conn.prepareStatement(qryActivities);
    int i = 0;
    for (final int id : subjectIDsByGroup) {
      stmtActivites.setInt(++i, id);
    }
    final ResultSet rsActivities = stmtActivites.executeQuery();
    final Map<Integer, FCAObject> activities = new HashMap<Integer, FCAObject>();
    while (rsActivities.next()) {
      final int id = rsActivities.getInt(1);
      double threshold = rsActivities.getDouble(4);
      if (rsActivities.wasNull())
        threshold = -Double.MAX_VALUE;
      activities.put(id, new FCAObject(rsActivities.getString(2), Double.toString(threshold), PREFIX_ACTIVITY + id));
    }
    return activities;
  }

  /**
   * Helper method getting all subject IDs of a group
   * 
   * @param gid
   *          the group's ID
   * @return the set of subject IDs
   * @throws SQLException
   */
  private Set<Integer> getSubjectIDsByGroup(final int gid) throws SQLException {
    final PreparedStatement stmt = conn.prepareStatement(MySQLConnector.STMT_SUBJECTS_BY_GROUP);
    stmt.setInt(1, gid);
    final ResultSet rsSubjectID = stmt.executeQuery();
    final Set<Integer> subjectIDs = new HashSet<>();
    while (rsSubjectID.next()) {
      subjectIDs.add(rsSubjectID.getInt(1));
    }
    return subjectIDs;
  }

  /**
   * Get all competences associated to a group
   * 
   * @param gid
   *          the group's ID
   * @return mapping of competences' MySQL DB IDs to competences as FCA Objects
   * @throws SQLException
   */
  public Map<Integer, FCAObject> getCompetenciesByGroup(final int gid) throws SQLException {
    final PreparedStatement stmt = conn.prepareStatement(MySQLConnector.STMT_COMPETENCIES_BY_GROUP);
    stmt.setInt(1, gid);
    final ResultSet rsCompentencies = stmt.executeQuery();
    final Map<Integer, FCAObject> competencies = new HashMap<>();
    while (rsCompentencies.next()) {
      competencies.put(rsCompentencies.getInt(1), new FCAObject(rsCompentencies.getString(2), "",
          MySQLConnector.PREFIX_COMPETENCE + rsCompentencies.getInt(1)));
    }
    return competencies;
  }

  /**
   * get all competences associated with a subject
   * 
   * @param subjectid
   *          the subject's ID
   * @return mapping of competences' MySQL DB IDs to competences as FCA Objects
   * @throws SQLException
   */
  public Map<Integer, FCAObject> getCompetenciesBySubject(final int subjectid) throws SQLException {
    final PreparedStatement stmt = conn
        .prepareStatement(MySQLConnector.buildStmtIN(MySQLConnector.STMT_COMPETENCIES_BY_SUBJECT, "#", 1));

    stmt.setInt(1, subjectid);

    final ResultSet rsCompentencies = stmt.executeQuery();
    final Map<Integer, FCAObject> competencies = new HashMap<>();
    int id = 0;
    while (rsCompentencies.next()) {
      id = rsCompentencies.getInt(2);
      if (competencies.containsKey(id)) {
        competencies.put(id, new FCAObject(rsCompentencies.getString(3), "",
            MySQLConnector.PREFIX_COMPETENCE + rsCompentencies.getInt(1)));
      }
    }
    return competencies;
  }

  /**
   * Get a mapping of students to (performed) activities of the specified group
   * 
   * @param gid
   *          the group's ID
   * @param activityIDsBySubject
   *          the MySQL IDs of the group's subjects
   * @return 3D matrix of students x activities x value of (student x activity)
   * @throws SQLException
   */
  public Map<String, Map<String, Double>> getActValuesByGroup(final int gid, final Set<Integer> activityIDsBySubject)
      throws SQLException {
    final Map<Integer, FCAAttribute> students = getStudents(gid);
    final String qry = MySQLConnector.buildActValueValueStmt(students.size(), activityIDsBySubject.size());
    final PreparedStatement stmt = conn.prepareStatement(qry);
    int i = 0;
    final Set<Integer> studentIDs = students.keySet();
    for (final int id : studentIDs) {
      stmt.setInt(++i, id);
    }
    for (final int id : activityIDsBySubject) {
      stmt.setInt(++i, id);
    }

    final ResultSet rsAct = stmt.executeQuery();
    final Map<String, Map<String, Double>> result = new HashMap<>();
    while (rsAct.next()) {
      final String studId = MySQLConnector.PREFIX_STUDENT + rsAct.getInt(1);
      final String actId = MySQLConnector.PREFIX_ACTIVITY + rsAct.getInt(2);
      double val = rsAct.getInt(3);
      if (rsAct.wasNull())
        val = -Double.MAX_VALUE;
      Map<String, Double> value;
      if (!result.containsKey(studId)) {
        value = new HashMap<>();
      } else {
        value = result.get(studId);
      }
      value.put(actId, val);
      result.put(studId, value);
    }
    return result;
  }

  /**
   * Get a mapping of students to (achieved) competences of the specified group
   * 
   * @param studentIds
   *          the students' MySQL IDs
   * @param compIds
   *          the competences' MySQL IDs
   * @return 3D matrix of students x competences x value of (student x competence)
   * @throws SQLException
   */
  public Map<String, Map<String, Double>> getCompValuesByStudents(final Set<Integer> studentIds,
      final Set<Integer> compIds) throws SQLException {
    final String qry = MySQLConnector.buildCompValueStmt(studentIds.size(), compIds.size());
    final PreparedStatement stmt = conn.prepareStatement(qry);
    int i = 0;
    for (final int id : studentIds) {
      stmt.setInt(++i, id);
    }
    for (final int id : compIds) {
      stmt.setInt(++i, id);
    }
    final ResultSet rsComp = stmt.executeQuery();
    final Map<String, Map<String, Double>> result = new HashMap<>();
    while (rsComp.next()) {
      final String studId = MySQLConnector.PREFIX_STUDENT + rsComp.getInt(1);
      final String compId = MySQLConnector.PREFIX_COMPETENCE + rsComp.getInt(2);
      double val = rsComp.getDouble(3);
      if (rsComp.wasNull())
        val = -Double.MIN_VALUE;
      Map<String, Double> value;
      if (!result.containsKey(studId)) {
        value = new HashMap<>();
      } else {
        value = result.get(studId);
      }
      value.put(compId, val);
      result.put(studId, value);
    }
    return result;
  }

  /**
   * Get a mapping of a group's competences to activities as specified in the MySQL Database
   * 
   * @param gid
   *          the group's ID
   * @return 2D matrix of competences x activities
   * @throws SQLException
   */
  public Map<FCAObject, Set<FCAAttribute>> getActivitiesToCompetences(final int gid) throws SQLException {
    final Map<Integer, FCAObject> competences = getCompetenciesByGroup(gid);
    final Set<Integer> subjectIDsByGroup = getSubjectIDsByGroup(gid);
    final String qry = MySQLConnector.buildStmtIN(MySQLConnector.STMT_ACTIVITIES_BY_COMPETENCES, "#",
        subjectIDsByGroup.size());
    final PreparedStatement stmt = conn.prepareStatement(qry);
    int i = 0;
    for (final int id : subjectIDsByGroup) {
      stmt.setInt(++i, id);
    }

    final ResultSet rsComp = stmt.executeQuery();
    final Map<Integer, FCAAttribute> activities = new HashMap<>();
    final Map<FCAObject, Set<FCAAttribute>> result = new HashMap<>();
    while (rsComp.next()) {
      final int activityID = rsComp.getInt(1);
      FCAAttribute activity;
      if (activities.containsKey(activityID)) {
        activity = activities.get(activityID);
      } else {
        activity = new FCAAttribute(rsComp.getString(2), "", MySQLConnector.PREFIX_ACTIVITY + activityID);
        activities.put(activityID, activity);
      }
      final FCAObject competence = competences.get(rsComp.getInt(3));
      if (!result.containsKey(competence)) {
        result.put(competence, new HashSet<FCAAttribute>());
      }
      result.get(competence).add(activity);
    }
    return result;
  }

  // Helper methods to construct set-based query statements
  private static String buildCompValueStmt(final int studCount, final int compCount) {
    return MySQLConnector.buildStmtIN(MySQLConnector.buildStmtIN(MySQLConnector.STMT_COMPLOG, "$", studCount), "#",
        compCount);
  }

  private static String buildActValueValueStmt(final int studCount, final int actCount) {
    return MySQLConnector.buildStmtIN(MySQLConnector.buildStmtIN(MySQLConnector.STMT_ACTLOG, "$", studCount), "#",
        actCount);
  }

  private static String buildStmtIN(final String stmt, final String placeholder, final int count) {
    final StringBuilder bld = new StringBuilder();
    final String s = "?,";
    for (int i = 0; i < count; ++i) {
      bld.append(s);
    }
    bld.setLength(bld.length() - 1);
    return stmt.replace(placeholder, bld.toString());
  }

  /**
   * for some basic testing and debugging purposes
   */
  public static void main(final String[] args) throws ClassNotFoundException, SQLException, IOException {

    final MySQLConnector db = new MySQLConnector();
    final int gid = 23;
    final Set<Integer> subjectIDsByGroup = db.getSubjectIDsByGroup(gid);
    System.out.print("SubjectIDS:");
    for (final int i : subjectIDsByGroup) {
      System.out.print(" " + i);
    }
    System.out.println();
    final Map<Integer, FCAAttribute> students = db.getStudents(gid);
    final String gName = db.getGroupName(gid);
    final Map<Integer, FCAObject> competences = db.getCompetenciesByGroup(gid);
    System.out.println("Group: " + gName);
    System.out.println("Competences:");
    for (final int i : competences.keySet()) {
      System.out.print(" " + i + ": " + competences.get(i) + ",");
    }
    System.out.println("\nStudents: ");
    for (final int i : students.keySet()) {
      System.out.print(i + " " + students.get(i).getName() + ",");
    }
    System.out.println("\nValues:");
    final Map<String, Map<String, Double>> compVals = db.getCompValuesByStudents(students.keySet(),
        competences.keySet());
    for (final String i : compVals.keySet()) {
      System.out.println(
          i + ", " + students.get(Integer.parseInt(i.substring(MySQLConnector.PREFIX_STUDENT.length()))) + ":");
      for (final String h : compVals.get(i).keySet()) {
        System.out.print("(" + competences.get(Integer.parseInt(h.substring(MySQLConnector.PREFIX_COMPETENCE.length())))
            + "," + compVals.get(i).get(h) + ")");
      }
      System.out.println();
    }

    System.out.println("Activities");
    final Map<String, Map<String, Double>> actVals = db.getActValuesByGroup(gid, db.getActivites(gid).keySet());
    for (final String i : actVals.keySet()) {
      System.out.println(
          i + ", " + students.get(Integer.parseInt(i.substring(MySQLConnector.PREFIX_STUDENT.length()))) + ":");
      for (final String h : actVals.get(i).keySet()) {
        System.out.println(h + ": " + actVals.get(i).get(h));
      }
      System.out.println();
    }

    System.out.println("Activites to competences");
    final Map<FCAObject, Set<FCAAttribute>> activitiesToCompetences = db.getActivitiesToCompetences(gid);
    for (final FCAObject o : activitiesToCompetences.keySet()) {
      System.out.println(o);
      for (final FCAAttribute a : activitiesToCompetences.get(o)) {
        System.out.println("\t" + a);
      }
    }

    // final DomainWrapper domain = new FCAService().createDomain(gid, 0.3);
    // System.out.println(domain);
  }
}
