package at.tugraz.kmi.medokyservice.fca.rest.wrappers;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Set;

import at.tugraz.kmi.medokyservice.fca.db.DataObject;
import at.tugraz.kmi.medokyservice.fca.db.domainmodel.Concept;
import at.tugraz.kmi.medokyservice.fca.db.domainmodel.FCAAttribute;
import at.tugraz.kmi.medokyservice.fca.db.domainmodel.FCAObject;
import at.tugraz.kmi.medokyservice.fca.db.domainmodel.LearningObject;

/**
 * Wrapper class resembling a {@link Concept} or a {@link LearnerConcept} but without cyclic or nested references used
 * by {@link LatticeWrapper}.
 * 
 * @author Bernd Prünster <mail@berndpruenster.org>
 * @see LatticeWrapper
 * @see Concept
 */
public class ConceptWrapper extends AbstractWrapper {
  /**
   * ID of the domain the concept belongs to
   */
  public long domainId;

  /**
   * flag indicating whether the concept is part of the taxonomy
   */
  public boolean partOfTaxonomy;

  /**
   * flag indicating whether the concept is an object concept
   */
  public boolean objectConcept;

  /**
   * flag indicating whether the concept is an attribute concept
   */
  public boolean attributeConcept;

  /**
   * attributes with valuations mapped to them
   */
  public LinkedHashMap<FCAAttribute, Float> attributes;

  /**
   * objects with valuations mapped to them
   */
  public LinkedHashMap<FCAObject, Float> objects;

  /**
   * @see Concept
   */
  public LinkedHashSet<FCAObject> uniqueObjects;

  /**
   * @see Concept
   */
  public LinkedHashSet<FCAAttribute> uniqueAttributes;

  /**
   * @see Concept
   */
  public HashSet<ConceptWrapper> successors;

  /**
   * @see Concept
   */
  public HashSet<ConceptWrapper> taxonomySuccessors;

  public Set<Long> clickedLearningObjects;

  public ConceptWrapper() {
  }
  /*
   * public <E extends DataObject> ConceptWrapper(E concept) { super.id = concept.getId(); super.name =
   * concept.getName(); super.description = concept.getDescription(); if (concept instanceof Concept) init((Concept)
   * concept); else if (concept instanceof LearnerConcept) init((LearnerConcept) concept); }
   */

  public <E extends DataObject> ConceptWrapper(E concept, Set<LearningObject> clickedLearningObjects) {
    super.id = concept.getId();
    super.name = concept.getName();
    super.description = concept.getDescription();
    if (concept instanceof Concept)
      init((Concept) concept);

  }

  @SuppressWarnings({ "unchecked", "rawtypes" })
  private void init(Concept concept) {

    partOfTaxonomy = concept.isPartOfTaxonomy();
    objectConcept = concept.isObjectConcept();
    attributeConcept = concept.isAttributeConcept();

    attributes = new LinkedHashMap<FCAAttribute, Float>();
    for (Comparable a : concept.getAttributes()) {
      attributes.put((FCAAttribute) a, 0.0f);
    }

    objects = new LinkedHashMap<FCAObject, Float>();
    for (Comparable o : concept.getObjects()) {
      objects.put((FCAObject) o, 0.0f);
    }
    uniqueAttributes = new LinkedHashSet<FCAAttribute>();
    uniqueAttributes.addAll((Collection<? extends FCAAttribute>) concept.getUniqueAttributes());

    uniqueObjects = new LinkedHashSet<FCAObject>();
    uniqueObjects.addAll((Collection<? extends FCAObject>) concept.getUniqueObjects());

    successors = new HashSet<ConceptWrapper>();
    taxonomySuccessors = new HashSet<ConceptWrapper>();
    clickedLearningObjects = new HashSet<Long>();

  }

  /**
   * Creates references successor Concepts
   * 
   * @param <E>
   * 
   * @param successors
   *          this concept's successors
   * @param taxonomySuccessors
   *          successors also part of the taxonomy
   */
  <E extends DataObject> void setSucessors(Set<E> successors, Set<E> taxonomySuccessors,
      Set<LearningObject> clickedLearningObjects) {
    for (E s : successors) {
      this.successors.add(new ConceptWrapper(s, clickedLearningObjects));
    }
    for (E s : taxonomySuccessors) {
      this.taxonomySuccessors.add(new ConceptWrapper(s, clickedLearningObjects));
    }
  }

}
